/*! \file   SplayTree.hpp
 *  \brief  A class for representing SplayTrees in a manner like that of the standard library.
 *  \author Peter Chapin <peter.chapin@vermontstate.edu>
 *
 * This file is a starter file for Homework #5 in CIS-3012, C++ Programming. Implement all
 * methods below marked with "Finish Me!" You do *not* have to deal with TODO items; they are
 * there only for future reference.
 */

#ifndef SPICA_SPLAYTREE_HPP
#define SPICA_SPLAYTREE_HPP

#include <cassert>
#include <cstddef>
#include <functional>
#include <initializer_list>
#include <memory>
#include <sstream>
#include <vector>

namespace spica {

    template<typename T, typename StrictWeakOrdering = std::less<T>>
    class SplayTree {
    private:
        struct Node;
        using NodePointer = std::shared_ptr<Node>;
        using WeakNodePointer = std::weak_ptr<Node>;

    public:
        //! The usual type aliases.
        using size_type       = std::size_t;
        using difference_type = std::ptrdiff_t;
        using value_type      = T;
        using reference       = const T &;
        using const_reference = const T &;
        using pointer         = const T *;
        using const_pointer   = const T *;

        //! An exception class thrown if a method is not implemented.
        class NotImplemented : public std::logic_error {
        public:
            NotImplemented( const std::string &message ) :
                std::logic_error( message )
            { }
        };

        //! Default constructor.
        SplayTree( StrictWeakOrdering swo = StrictWeakOrdering( ) ) :
            root( NodePointer( ) ), compare( swo ), node_count( 0 )
        { }

        //! Destructor (generated destructor is fine)
        // ~SplayTree( );

        //! Copy constructor (deleted for now).
        SplayTree( const SplayTree &other ) = delete;

        //! Copy assignment operator (deleted for now).
        SplayTree &operator=( const SplayTree &other ) = delete;

        //! Move constructor.
        SplayTree( SplayTree &&other );

        //! Move assignment operator.
        SplayTree &operator=( SplayTree &&other );

        //! Initializer list constructor.
        SplayTree( std::initializer_list<T> init_list, StrictWeakOrdering swo = StrictWeakOrdering( ) );

        //! Iterator class.
        class iterator {
        public:
            // The usual type aliases.
            using iterator_category = std::bidirectional_iterator_tag;
            using value_type        = T;
            using difference_type   = std::ptrdiff_t;
            using pointer           = const T *;
            using reference         = const T &;

            iterator( const SplayTree *t ) :
                tree( t ), current( NodePointer( ) )
                { }
            iterator( const SplayTree *t, NodePointer ptr ) :
                tree( t ), current( ptr )
                { }
            
            const T &operator*( ) const { return current->data; }
            const T *operator->( ) const { return &current->data; }

            iterator &operator++( );     // Prefix version.
            iterator operator++( int );  // Postfix version.
            iterator &operator--( );     // Prefix version.
            iterator operator--( int );  // Postfix version.

            bool operator==( const iterator &other ) const { return current == other.current; }
            bool operator!=( const iterator &other ) const { return current != other.current; }

        private:
            const SplayTree  *tree;      // This is needed so end( ) is decrementable.
            NodePointer current;
        };  // End of iterator class.

        iterator begin( ) const;
        iterator end( ) const { return iterator( this ); }

        //! Insert
        std::pair<iterator, bool> insert( const T &value );
        template<typename InputIterator>
            void insert( InputIterator first, InputIterator last );
        void insert( std::initializer_list<T> init_list );

        //! Find
        iterator find( const T &value );

        //! Erase
        size_type erase( const T &value );
        iterator erase( iterator position );

        // Testing/Debugging
        // -----------------

        //! An exception class thrown `check_structure( )` detects a problem with the tree structure.
        /*!
         * If this occurs, the code is buggy and undefined behavior is likely.
         */
        class InconsistentStructure : public std::logic_error {
        public:
            InconsistentStructure( const std::string &message ) :
                std::logic_error( message )
            { }
        };
        void check_structure( ) const;

        struct DumpItem {
            size_type depth;
            T data;

            bool operator==( const DumpItem &other ) const
            {
                return depth == other.depth && data == other.data;
            }
        };
        using DumpResult = std::vector<DumpItem>;
        DumpResult dump( ) const;

    private:
        struct Node {
            T data;
            WeakNodePointer parent;  // A weak pointer to break reference cycles.
            NodePointer left;
            NodePointer right;

            Node( const T &d,
                  WeakNodePointer p = WeakNodePointer( ),
                  NodePointer l = NodePointer( ),
                  NodePointer r = NodePointer( ) ) :
                data( d ), parent( p ), left( l ), right( r )
            { }
        };

        NodePointer root;
        StrictWeakOrdering compare;
        size_type node_count;

        // The rotations take a pointer to the node initially at the root of the subtree. They
        // rotate that node down and toward the direction indicated by the name of the function.
        //
        void rotate_left( NodePointer x );
        void rotate_right( NodePointer y );

        // Starting at node x, splay the tree until x is the root.
        void splay( NodePointer x );

        // Find the node with the minimum (maximum) value in the subtree rooted at subtree_root.
        static NodePointer minimum_node( NodePointer subtree_root );
        static NodePointer maximum_node( NodePointer subtree_root );

        // Testing/Debugging
        // -----------------
        size_type traverse_check( NodePointer p ) const;
        std::vector<DumpItem> traverse_dump( NodePointer p, size_type depth ) const;
    };

    // Implementation
    // ==============

    template<typename T, typename StrictWeakOrdering>
    SplayTree<T, StrictWeakOrdering>::SplayTree( SplayTree &&other )
    {
        root       = std::move( other.root );
        compare    = other.compare;
        node_count = other.node_count;
        other.node_count = 0;
    }

    template<typename T, typename StrictWeakOrdering>
    SplayTree<T, StrictWeakOrdering> &SplayTree<T, StrictWeakOrdering>::operator=( SplayTree &&other )
    {
        if( this != &other ) {
            root       = std::move( other.root );
            compare    = other.compare;
            node_count = other.node_count;
            other.node_count = 0;
        }
        return *this;
    }

    template<typename T, typename StrictWeakOrdering>
    SplayTree<T, StrictWeakOrdering>::SplayTree( std::initializer_list<T> init_list, StrictWeakOrdering swo ) :
        root( NodePointer( ) ), compare( swo ), node_count( 0 )
    {
        for( const auto &item : init_list ) {
            insert( item );
        }
    }

    template<typename T, typename StrictWeakOrdering>
    typename SplayTree<T, StrictWeakOrdering>::iterator &
        SplayTree<T, StrictWeakOrdering>::iterator::operator++( )
    {
        throw NotImplemented( "iterator::operator++" );
        
        // Finish Me!
        //
        // IF <current is not NULL> THEN
        //   IF <current's right child exists> THEN
        //     <set current to the minimum node in the right subtree>
        //   ELSE
        //     WHILE <current has a parent AND current is its parent's right child> LOOP
        //       <set current to be its parent>
        //     END
        //     <set current to be its parent (which might be NULL)>
        //   END
        // END

        return *this;
    }

    template<typename T, typename StrictWeakOrdering>
    typename SplayTree<T, StrictWeakOrdering>::iterator
        SplayTree<T, StrictWeakOrdering>::iterator::operator++( int )
    {
        iterator saved{ *this };
        ++( *this );
        return saved;
    }

    template<typename T, typename StrictWeakOrdering>
    typename SplayTree<T, StrictWeakOrdering>::iterator &
        SplayTree<T, StrictWeakOrdering>::iterator::operator--( )
    {
        throw NotImplemented( "iterator::operator--" );

        // Finish Me!
        //
        // IF <we are an off-the-end iterator> THEN
        //   IF <the tree is not empty> THEN
        //     <set current to be the maximum node in the tree>
        //   END
        // ELSE
        //   <similar logic as for operator++ except with left and right reversed>
        // END
        
        return *this;
    }

    template<typename T, typename StrictWeakOrdering>
    typename SplayTree<T, StrictWeakOrdering>::iterator
        SplayTree<T, StrictWeakOrdering>::iterator::operator--( int )
    {
        iterator saved{ *this };
        --( *this );
        return saved;
    }

    template<typename T, typename StrictWeakOrdering>
    typename SplayTree<T, StrictWeakOrdering>::iterator
        SplayTree<T, StrictWeakOrdering>::begin( ) const
    {
        if( root == nullptr ) return end( );
        return iterator( this, minimum_node( root ) );
    }

    template<typename T, typename StrictWeakOrdering>
    std::pair<typename SplayTree<T, StrictWeakOrdering>::iterator, bool>
        SplayTree<T, StrictWeakOrdering>::insert( const T &value )
    {
        throw NotImplemented( "insert" );

        // Finish Me!

        NodePointer new_node = std::make_shared<Node>( value );

        // IF <root is NULL> THEN
        //   <install the new node as the root>
        //   return { iterator( this, root ), true };
        // END
        //
        // <set current to the root>
        // WHILE TRUE LOOP
        //   IF <value comes before the current node> THEN
        //     IF <the current node has no left child> THEN
        //       <install the new node as the left child of the current node>
        //       <splay the new node>
        //       return { iterator( this, new_node ), true };
        //     ELSE
        //       <move current to the left>
        //     END
        //   ELSE IF <the current node comes before value> THEN
        //     <similar to above, except to the right>
        //   ELSE
        //     // An equivalent value already exists in the three.
        //     // DO NOT splay in this case!
        //     return { iterator( this, current ), false };
        //   END
        // END

        // This code will never be reached. Replace with assert( false ) in your version.
        return { iterator( this, root ), false };
        
    }

    template<typename T, typename StrictWeakOrdering>
    template<typename InputIterator>
        void SplayTree<T, StrictWeakOrdering>::insert( InputIterator first, InputIterator last )
    {
        while( first != last ) {
            insert( *first );
            ++first;
        }
    }

    template<typename T, typename StrictWeakOrdering>
    void SplayTree<T, StrictWeakOrdering>::insert( std::initializer_list<T> init_list )
    {
        insert( init_list.begin( ), init_list.end( ) );
    }

    template<typename T, typename StrictWeakOrdering>
    typename SplayTree<T, StrictWeakOrdering>::iterator
        SplayTree<T, StrictWeakOrdering>::find( const T &value )
    {
        throw NotImplemented( "find" );

        // Finish Me!
        
        return iterator( this );
    }

    template<typename T, typename StrictWeakOrdering>
    typename SplayTree<T, StrictWeakOrdering>::size_type
        SplayTree<T, StrictWeakOrdering>::erase( const T &value )
    {
        // DO NOT IMPLEMENT!
        throw NotImplemented( "erase" );
        return 0;
    }

    template<typename T, typename StrictWeakOrdering>
    typename SplayTree<T, StrictWeakOrdering>::iterator
        SplayTree<T, StrictWeakOrdering>::erase( iterator position )
    {
        // DO NOT IMPLEMENT!
        throw NotImplemented( "erase" );
        return iterator( );
    }

    // Private Methods
    // ---------------

    template<typename T, typename StrictWeakOrdering>
    void SplayTree<T, StrictWeakOrdering>::rotate_left( NodePointer x )
    {
        assert( x != nullptr );

        // Turn y's left subtree into x's right subtree
        NodePointer y = x->right;
        x->right = y->left;

        // If y's left subtree is not empty, set its parent to x.
        if( y->left != nullptr ) {
            y->left->parent = x;
        }

        // Link x's parent to y.
        y->parent = x->parent;

        // If x was the root...
        if( x->parent.lock( ) == nullptr ) {
            root = y;
        }
        // ... otherwise fix x's parent to point at y.
        else if( x == x->parent.lock( )->left ) {
            x->parent.lock( )->left = y;
        }
        else {
            x->parent.lock( )->right = y;
        }

        // Attach x as the left child of y.
        y->left = x;
        x->parent = y;
    }

    template<typename T, typename StrictWeakOrdering>
    void SplayTree<T, StrictWeakOrdering>::rotate_right( NodePointer y )
    {
        assert( y != nullptr );

        throw NotImplemented( "rotate_right" );

        // Finish Me!

    }

    template<typename T, typename StrictWeakOrdering>
    void SplayTree<T, StrictWeakOrdering>::splay( NodePointer x )
    {
        assert( x != nullptr );

        throw NotImplemented( "splay" );

        // Finish Me!
        //
        // WHILE <x's parent is not NULL> LOOP
        //   // This handles the two cases when x is an immediate child of the root.
        //   IF <x's grandparent is NULL> THEN
        //     IF <x is its parent's left child> THEN
        //       <rotate x's parent to the right>
        //     ELSE
        //       <rotate x's parent to the left>
        //     END
        //
        //   // x has a grandparent.
        //   ELSE IF <x is its parent's left child> THEN
        //     IF <x's parent is the left child of x's grandparent> THEN
        //       // Zig-Zig
        //       <rotate x's grandparent to the right>
        //       <rotate x's parent to the right>
        //     ELSE
        //       // Zig-Zag
        //       <rotate x's parent to the right>
        //       <rotate x's *NEW* parent to the left>
        //     END
        //   ELSE
        //     // x is its parent's right child
        //     <similar logic to above>
        //   END
        // END

    }

    template<typename T, typename StrictWeakOrdering>
    typename SplayTree<T, StrictWeakOrdering>::NodePointer
        SplayTree<T, StrictWeakOrdering>::minimum_node( NodePointer subtree_root )
    {
        assert( subtree_root != nullptr );

        NodePointer current = subtree_root;
        while( current->left != nullptr ) {
            current = current->left;
        }
        return current;
    }

    template<typename T, typename StrictWeakOrdering>
    typename SplayTree<T, StrictWeakOrdering>::NodePointer
        SplayTree<T, StrictWeakOrdering>::maximum_node( NodePointer subtree_root )
    {
        assert( subtree_root != nullptr );

        NodePointer current = subtree_root;
        while( current->right != nullptr ) {
            current = current->right;
        }
        return current;
    }

    // Testing/Debugging
    // -----------------

    template<typename T, typename StrictWeakOrdering>
    void SplayTree<T, StrictWeakOrdering>::check_structure( ) const
    {
        if( root == nullptr && node_count == 0 ) return;
        if( root == nullptr && node_count != 0 )
            throw InconsistentStructure( "Non-zero node count with a null root" );
        // The root cannot be nullptr beyond this point.

        if( root->parent.lock( ) != nullptr )
            throw InconsistentStructure( "Root has a non-null parent" );

        // Explore the tree structure and count the number of nodes.
        size_type actual_count = traverse_check( root );
        if( node_count != actual_count ) {
            std::ostringstream formatter;
            formatter << "node_count (" << node_count << ") != actual_count (" << actual_count << ")";
            throw InconsistentStructure( formatter.str( ) );
        }
    }

    template<typename T, typename StrictWeakOrdering>
    typename SplayTree<T, StrictWeakOrdering>::size_type
        SplayTree<T, StrictWeakOrdering>::traverse_check( NodePointer p ) const
    {
        size_type subtree_count = 0;
        if( p != nullptr ) {
            subtree_count++;

            if( p->left != nullptr ) {
                // Note that this check is incomplete. *Everything* in the left subtree should
                // come before p->data. This only verifies that the root of the left subtree
                // satisfies the invariant. Conceivably, nodes to the right of the left child
                // should actually be to p's right. That won't be caught here.
                //
                if( !compare( p->left->data, p->data ) ) {
                    throw InconsistentStructure( "Left child out of order" );
                }
                if( p->left->parent.lock( ) != p ) {
                    throw InconsistentStructure( "Left child has bad parent" );
                }
                subtree_count += traverse_check( p->left );
            }
            if( p->right != nullptr ) {
                // See comment above. A mirror issue exists here.
                //
                if( !compare( p->data, p->right->data ) ) {
                    throw InconsistentStructure( "Right child out of order" );
                }
                if( p->right->parent.lock( ) != p ) {
                    throw InconsistentStructure( "Right child has bad parent" );
                }
                subtree_count += traverse_check( p->right );
            }
        }
        return subtree_count;
    }

    template<typename T, typename StrictWeakOrdering>
    std::vector<typename SplayTree<T, StrictWeakOrdering>::DumpItem>
        SplayTree<T, StrictWeakOrdering>::dump( ) const
    {
        return traverse_dump( root, 0 );
    }

    template<typename T, typename StrictWeakOrdering>
    typename SplayTree<T, StrictWeakOrdering>::DumpResult
        SplayTree<T, StrictWeakOrdering>::traverse_dump( NodePointer p, size_type depth ) const
    {
        DumpResult result;
        if( p != nullptr ) {
            auto left_result  = traverse_dump( p->left,  depth + 1 );
            result.insert( result.end( ), left_result.begin( ), left_result.end( ) );
            result.push_back( { depth, p->data } );
            auto right_result = traverse_dump( p->right, depth + 1 );
            result.insert( result.end( ), right_result.begin( ), right_result.end( ) );
        }
        return result;
    }

}

#endif
